; preload.lisp   by Michael Thorpe   2006-08-15

(#define '#f (= 1 2))
(#define '#t (= 1 1))
(#define 'pair? (#lambda '(o) '(eq? 'cons (object-type o))))
(#define 'stream? (#lambda '(o) '(eq? 'stream (object-type o))))
(#define 'list (#lambda 'aaa 'aaa))

(#define 'get
	(#lambda '(hash item)
		'(#ifelse hash
			'(#ifelse (eq? (car (car hash)) item)
				'(cdr (car hash))
				'(get (cdr hash) item))
			'#f)))

(#define 'macroexpansions '())
(#define 'dont-macro-expand '(dont . macro-expand))
(#define 'macro-expand
	(#lambda '(o)
		'(#ifelse (pair? o)
			'(#prog2of2
				(#let 'macro (get macroexpansions (car o))
					'(#ifelse macro
						'(#let 'cdr-o (cdr o)
							'(#let 'expansion (macro . cdr-o)
								'(#ifelse (eq? expansion dont-macro-expand)
									'#f
									'(#set 'o (macro-expand expansion)))))
						'#f))
				'(#map 'o '(macro-expand o) o))
			'o)))
(#define '#macro-define
	(#lambda '(macro func)
		'(#prog2of2
			(#set 'macroexpansions (cons (cons macro func) macroexpansions))
			'macro)))

(#macro-define 'define
	(#lambda '(name . value)
		'(#ifelse (pair? name)
			'(list '#define (quote (car name)) (list 'lambda (cdr name) (cons 'progn value)))
			'(list '#define (quote name) . value))))
(#macro-define 'lambda
	(#lambda '(args . prog)
		'(list '#lambda (quote args) (quote (macro-expand (cons 'progn prog))))))
(#macro-define 'let
	(#lambda '(vars . prog)
		'(#ifelse (pair? vars)
			'(list
				'#let
				(quote (car (car vars)))
				(cons 'progn (cdr (car vars)))
				(quote (macro-expand (cons 'let (cons (cdr vars) prog)))))
			'(cons 'progn prog))))
(#macro-define 'macro-define
	(#lambda '(name . value)
		'(#ifelse (pair? name)
			'(list '#macro-define (quote (car name)) (list 'lambda (cdr name) (cons 'progn value)))
			'(list '#macro-define (quote name) . value))))
(#macro-define 'progn
	(#lambda 'aaa
		'(#ifelse aaa
			'(#ifelse (pair? aaa)
				'(#ifelse (pair? (cdr aaa))
					'(list '#prog2of2 (car aaa) (quote (macro-expand (cons 'progn (cdr aaa)))))
					'(car aaa))
				'aaa)
			'aaa)))
(#macro-define 'set
	(#lambda '(name . value)
		'(#ifelse (pair? name)
			'(list '#set (quote (car name)) (list 'lambda (cdr name) (cons 'progn value)))
			'(list '#set (quote name) . value))))

(#macro-define 'catch
	(#lambda '(cond body handler)
		'(list '#catch cond
			(quote (macro-expand body))
			(quote (macro-expand handler)))))
(#macro-define 'ifelse
	(#lambda '(cond iftrue iffalse)
		'(list '#ifelse cond (quote (macro-expand iftrue)) (quote (macro-expand iffalse)))))
(#macro-define 'loop
	(#lambda 'body
		'(list '#loop (quote (macro-expand (cons 'progn body))))))
(#macro-define 'map
	(#lambda '(var mapper . args)
		'(list '#map (quote var) (quote (macro-expand mapper)) . args)))

(#define 'evaluate-file
	(#lambda '(file)
		'(#catch 'end-of-file
			'(#loop '(evaluate (macro-expand (read-object file))))
			'#f)))

(evaluate-file #preload.lisp)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; From this point on we're operating on our own repl (minus the printing) ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro-define (+ . aaa)
	(ifelse (pair? (cdr aaa))
		(ifelse (pair? (cdr (cdr aaa)))
			(list '+ (car aaa) (cons '+ (cdr aaa)))
			dont-macro-expand)
		(car aaa)))
(macro-define (- . aaa)
	(ifelse (pair? (cdr aaa))
		(ifelse (pair? (cdr (cdr aaa)))
			(list '- (car aaa) (cons '+ (cdr aaa)))
			dont-macro-expand)
		(car aaa)))
(macro-define (* . aaa)
	(ifelse (pair? (cdr aaa))
		(ifelse (pair? (cdr (cdr aaa)))
			(list '* (car aaa) (cons '* (cdr aaa)))
			dont-macro-expand)
		(car aaa)))
(macro-define (/ . aaa)
	(ifelse (pair? (cdr aaa))
		(ifelse (pair? (cdr (cdr aaa)))
			(list '/ (car aaa) (cons '* (cdr aaa)))
			dont-macro-expand)
		(car aaa)))
(macro-define (and a . b)
	(ifelse b
		(list 'ifelse a (cons 'and b) #f)
		a))
(macro-define (defined? word)
	(ifelse (eq? 'word (object-type word))
		(list 'catch ''undefined
			(list 'progn word #t)
			#f)
		(throw 'type-error word)))
(macro-define (exit-loop . value)
	(list 'throw ''exit-loop (cons 'progn value)))
(macro-define (if bool . progs)
	(list 'ifelse bool
		(cons 'progn progs)
		#f))
(define (member item list)
	(ifelse list
		(ifelse (pair? list)
			(ifelse (eq? item (car list))
				list
				(member item (cdr list)))
			(ifelse (eq? item list)
				list
				#f))
		#f))
(macro-define (next-loop) (list 'throw ''next-loop #f))
(macro-define (not a) (list 'ifelse a #f #t))
(macro-define (or a . b)
	(ifelse b
		(let ((var (unique-word)))
			(list 'let
				(list (list var a))
				(list 'ifelse var var (cons 'or b))))
		a))
(macro-define (unless bool . body)
	(list 'ifelse bool
		#f
		(cons 'progn body)))
(macro-define (while cond . body)
	(list 'loop
		(list 'ifelse cond
			(cons 'progn body)
			(list 'throw ''exit-loop #f))))

(define (reverse/append straight newtail)
	(while straight
		(set newtail (cons (car straight) newtail))
		(set straight (cdr straight)))
	newtail)
(define (reverse straight)
	(reverse/append straight '()))
(define (append oldlist newtail)
	(reverse (reverse/append oldlist newtail)))

(define (write-newline file)
	(write-char file 10))
(define (write-object file obj)
	(write-string file (->string obj)))
(define (dprint . o)
	(while o
		(write-object #standard-out (car o))
		(set o (cdr o))
		(if o
			(write-char #standard-out 32)))
	(write-newline #standard-out)
	#f)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Now we reimplement the reader and read-object ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; This object is placed in the CAR of a cons to indicate that quasiquote
; should replace the cons with the CDR of the cons (but not quoted).
; For example, if read-object is about to read:
;   ,bar
; it will call macro-comma, which calls read-object, gets bar, and returns:
;   ('(unquote . identifier) . bar)
;
; If read-object is called where the stream starts with:
;   `(foo ,bar baz)
; then the backtick will be read, macro-backtick will call read-object and
; pass the following to quasiquote:
;   (foo ('(unquote . identifier) . bar) baz)
; Then quasiquote will do its thing and return:
;   (list 'foo bar 'baz)
;
; Note that quasiquote needs to be able to handle things like:
;   `(foo bar . ,(car baz))
; turning them into:
;   (cons 'foo (cons 'bar (car baz)))
;
(define unquote-identifier '(unquote . identifier))

(define character-macros '())
(define (define-character-macro char macro)
	(set character-macros (cons (cons char macro) character-macros)))

(define (get-using-= hash item)
	(while hash
		(if (= (car (car hash)) item)
			(exit-loop (cdr (car hash))))
		(set hash (cdr hash))))

(define (expand-string oldstring)
	(let ((i 0)
	      (newstring (new-string (+ 512 (string-length oldstring)))))
		(while (< i (string-length oldstring))
			(put-char newstring i (get-char oldstring i))))
	newstring)

(define (macro-singlequote stream)
	(catch 'end-of-file
		(quote (read-object stream))
		(throw 'reader-error stream)))
(define-character-macro 39 macro-singlequote)

; FIXME: We should use peek-char so the stream will point to the offending
; token if it's not a close parenthesis
(define (submacro-dot stream)
	(catch 'end-of-file
		(let ((o (read-object stream)))
			(ifelse (not (eq? ') (read-object stream)))
				(throw 'reader-error stream)
				o))
		(throw 'reader-error stream)))

(define (macro-openparen stream)
	(let ((reverse-list '())
	      (new-list '()))
		(catch 'end-of-file
			(loop
				(let ((o (read-object stream)))
					(ifelse (eq? o '))
						(exit-loop)
						(ifelse (eq? o '.)
							(ifelse reverse-list
								(progn
									(set new-list (submacro-dot stream))
									(exit-loop))
								(throw 'reader-error stream))
							(set reverse-list (cons o reverse-list))))))
			(throw 'reader-error stream))
		(reverse/append reverse-list new-list)))
(define-character-macro 40 macro-openparen)

(define (macro-closeparen stream)
	'))
(define-character-macro 41 macro-closeparen)

(define (macro-comma stream)
	(catch 'end-of-file
		(cons unquote-identifier (read-object stream))
		(throw 'reader-error stream)))
(define-character-macro 44 macro-comma)

(define (macro-semicolon stream)
	(let ((c 0))
		(while (and (<> c 10) (<> c 13))
			(set c (read-char stream))))
	(read-object stream))
(define-character-macro 59 macro-semicolon)

(define (quasiquote obj)
	(ifelse obj
		(ifelse (pair? obj)
			(ifelse (eq? unquote-identifier (car obj))
				(cdr obj)
				(ifelse (member unquote-identifier obj)
					(list 'cons (quasiquote (car obj)) (quasiquote (cdr obj)))
					(cons 'list (map obj (quasiquote obj) obj))))
			(quote obj))
		obj))

(define (macro-backtick stream)
	(quasiquote
		(catch 'end-of-file
			(read-object stream)
			(throw 'reader-error stream))))
(define-character-macro 96 macro-backtick)

(define (macro-doublequote stream)
	(let ((c 0)
	      (i 0)
	      (string (new-string 512)))
		(loop
			(if (= i (string-length string))
				(set string (expand-string string)))
			(set c
				(catch 'end-of-file
					(read-char stream)
					(throw 'reader-error stream)))
			(if (= c 34) ; 34 = ASCII double quote
				(exit-loop))
			(if (= c 92) ; 92 = ASCII backslash
				(set c 
					(catch 'end-of-file
						(read-char stream)
						(throw 'reader-error stream))))
			(put-char string i c)
			(set i (+ i 1)))
		(substring string 0 i)))
(define-character-macro 34 macro-doublequote)

(define (read-token stream c)
	(let ((buffer (new-string 512))
	      (i 0))
		(loop
			(if (= i (string-length buffer))
				(set buffer (expand-string buffer)))
			(put-char buffer i c)
			(set i (+ i 1))
			(catch 'end-of-file
				(set c (peek-char stream))
				(exit-loop))
			(if (isspace c)
				(exit-loop))
			(unless (isascii c)
				(exit-loop))
			(if (get-using-= character-macros c)
				(exit-loop))
			(set c (read-char stream)))
		(set buffer (substring buffer 0 i))
		(catch 'type-error
			(string->integer buffer)
			(string->symbol buffer))))

(set (read-object stream)
	(let ((c 32) ; 32 = ASCII space
	      (macro #f))
		(while (isspace c)
			(set c (read-char stream)))
		(unless (isascii c)
			(throw 'reader-error stream))
		(set macro (get-using-= character-macros c))
		(ifelse macro
			(macro stream)
			(read-token stream c))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Some more utility routines ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(macro-define (caar o) `(car (car ,o)))
(macro-define (cadr o) `(car (cdr ,o)))
(macro-define (cdar o) `(cdr (car ,o)))
(macro-define (cddr o) `(cdr (cdr ,o)))
(macro-define (caaar o) `(car (car (car ,o))))
(macro-define (caadr o) `(car (car (cdr ,o))))
(macro-define (cadar o) `(car (cdr (car ,o))))
(macro-define (caddr o) `(car (cdr (cdr ,o))))
(macro-define (cdaar o) `(cdr (car (car ,o))))
(macro-define (cdadr o) `(cdr (car (cdr ,o))))
(macro-define (cddar o) `(cdr (cdr (car ,o))))
(macro-define (cdddr o) `(cdr (cdr (cdr ,o))))

(define (cars arg) ; ((1 2) (3 4)) => (1 3)
	(let ((stack '()))
		(while arg
			(set stack (cons (caar arg) stack))
			(set arg (cdr arg)))
		(reverse stack)))
(define (cdrs arg) ; ((1 2) (3 4)) => ((2) (4))
	(let ((stack '()))
		(while arg
			(set stack (cons (cdar arg) stack))
			(set arg (cdr arg)))
		(reverse stack)))
(define (cadrs arg) ; ((1 2) (3 4)) => (2 4)
	(let ((stack '()))
		(while arg
			(set stack (cons (cadar arg) stack))
			(set arg (cdr arg)))
		(reverse stack)))

;
; my let@ is like Lisp's let, whereas my let is like Lisp's let*.
;
(macro-define (let@ vars . body)
	(ifelse vars
		(ifelse (cdr vars)
			(let ((varnames (reverse (cdr (cars vars))))
			      (firstvar (caar vars)))
				(set body (list 'progn `(set ,firstvar (car ,firstvar)) . body))
				(while varnames
					(set body
						`(let ((,(car varnames) (car ,firstvar)))
							(set ,firstvar (cdr ,firstvar))
							,body))
					(set varnames (cdr varnames)))
				`(let
					((,firstvar ,(cons 'list (reverse (cadrs vars)))))
					,body))
			(list 'let vars . body))
		(list 'progn . body)))

(define (string-eqv? a b)
	(let ((c (string-length a)))
		(ifelse (= c (string-length b))
			(loop
				(if (= c 0)
					(exit-loop #t))
				(set c (- c 1))
				(if (<> (get-char a c) (get-char b c))
					(exit-loop #f)))
			#f)))
(define (eqv? a b)
	(let ((ot-a (object-type a)))
		(ifelse (eq? a b)
			#t
			(ifelse (not (eq? ot-a (object-type b)))
				#f
				(ifelse (eq? ot-a 'integer)
					(= a b)
					(ifelse (eq? ot-a 'quote)
						(eqv? (evaluate a) (evaluate b))
						(ifelse (eq? ot-a 'string)
							(string-eqv? a b)
							(ifelse (eq? ot-a 'cons)
								(and
									(eqv? (car a) (car b))
									(eqv? (cdr a) (cdr b)))
								#f))))))))

;
; switch (takes the place of case)
;
; Format:
;  (switch key
;  	((value1) stuff1 stuff2)
;  	((value2 value3) stuff3 stuff4 stuff5)
;  	(else stuff6 stuff7))
; else is required to be the last case (and defaults to #f if not present)
;
(define (switch-handle var cases)
	(ifelse cases
		(ifelse (eq? 'else (caar cases))
			(cons 'progn (cdar cases))
			(ifelse (caar cases)
				`(ifelse (eqv? ,var ,(caaar cases))
					,(cons 'progn (cdar cases))
					,(switch-handle var
						(cons
							(cons (cdaar cases) (cdar cases))
							(cdr cases))))
				(switch-handle var (cdr cases))))
		#f))
(macro-define (switch key . cases)
	(let ((var (unique-word)))
		`(progn
			(let ((,var ,key))
				,(switch-handle var cases)))))

;
; bind operates similar to the PostScript bind operator.  It recursively
; looks up all words and substitutes the values of the words it finds.
;
(define (isinlist item list)
	(ifelse (pair? list)
		(ifelse (eq? item (car list))
			#t
			(isinlist item (cdr list)))
		(eq? item list)))
(define (bind bind#obj . bind#nowords)
	(switch (object-type bind#obj)
		(('cons) (map a (bind a . bind#nowords) bind#obj))
		(('func)
			(set bind#nowords (args-of-func bind#obj))
			(#lambda bind#nowords (bind (prog-of-func bind#obj) . bind#nowords)))
		(('quote) (quote (bind (evaluate bind#obj) . bind#nowords)))
		(('word)
			(ifelse (isinlist bind#obj bind#nowords)
				bind#obj
				(catch 'undefined (evaluate bind#obj) bind#obj)))
		(else bind#obj)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; And here's our repl ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (load-file filename)
	(evaluate-file (open-file filename "r")))

(define (show-error error errorvalue)
	(write-newline #standard-out)
	(write-string #standard-out "EXCEPTION: ")
	(write-object #standard-out error)
	(write-newline #standard-out)
	(write-string #standard-out "EXCEPTION-VALUE: ")
	(write-object #standard-out errorvalue)
	(write-newline #standard-out))
(define (read-eval-print in out prompt)
	(let ((scratchspace #f))
		(loop
			(if prompt (write-string out prompt))
			(set scratchspace
				(catch 'end-of-file
					(read-object in)
					(progn (if prompt (write-newline out)) (exit-loop))))
			(set scratchspace
				(catch #f
					(evaluate (macro-expand scratchspace))
					(progn (show-error #exception #exceptionvalue) (ifelse prompt (next-loop) (exit-loop)))))
			(if (and prompt scratchspace)
				(write-object out scratchspace)
				(write-newline out)))))
(read-eval-print #standard-in #standard-out (ifelse (isatty #standard-in) "mikelisp> " #f))
